@@ -4,10 +4,9 @@
 maniphono __init__.py
 """
 
-# Version of the lpngram package
-__version__ = "0.2.1"
+__version__ = "0.3.1"
 __author__ = "Tiago Tresoldi"
-__email__ = "tresoldi@shh.mpg.de"
+__email__ = "tiago.tresoldi@lingfil.uu.se"
 
 # Build the namespace
 from maniphono.phonomodel import (
@@ -16,10 +15,17 @@ from maniphono.phonomodel import (
     model_tresoldi,
 )  # pyflakes.ignore
 from maniphono.sound import Sound  # noqa: F401
-from maniphono.segment import Segment  # noqa: F401
-from maniphono.sequence import Sequence  # noqa: F401
+from maniphono.segment import (
+    Segment,
+    SoundSegment,
+    BoundarySegment,
+    parse_segment,
+)  # noqa: F401
+from maniphono.segsequence import SegSequence, parse_sequence  # noqa: F401
 from maniphono.utils import (
     codepoint2glyph,
     replace_codepoints,
     read_distance_matrix,
+    parse_constraints,
+    split_fvalues_str,
 )  # pyflakes.ignore
