@@ -2,10 +2,11 @@
 Utility functions for `maniphono`.
 """
 
-from pathlib import Path
-import re
 import csv
+import re
 import unicodedata
+from pathlib import Path
+from typing import List, Optional, Tuple
 
 # Pattern for unicode codepoint replacement
 RE_CODEPOINT = re.compile(r"[Uu]\+[0-9A-Fa-f]{4}")
@@ -15,32 +16,31 @@ RE_FEATURE = re.compile(r"^[a-z][-_a-z]*$")
 RE_FVALUE = re.compile(r"^[a-z][-_a-z]*$")
 
 
-def normalize(grapheme):
+def normalize(grapheme: str) -> str:
     """
     Normalize the string representation of a grapheme.
 
-    Currently, normalization involves NFD Unicode normalization and stripping
-    any leading and trailing white spaces. As these might change in the future, it
-    is suggested to always use this functions instead of reimplementing it in code.
+    Currently, normalization involves NFD Unicode normalization and stripping any leading
+    and trailing white spaces. As these might change in the future, it is suggested to
+    always use this function instead of reimplementing it in code each time.
 
-    Parameters
-    ----------
-    grapheme : str
-        The grapheme to be normalized.
-
-    Return
-    ------
-    grapheme : str
-        The normalized version of the grapheme.
+    @param grapheme: The grapheme to be normalized.
+    @return: The normalized version of the grapheme.
     """
 
     return unicodedata.normalize("NFD", grapheme).strip()
 
 
-# TODO: extend documentation
-def parse_constraints(constraints_str):
+# TODO: comment on the syntax that is accepted
+# TODO: annotate the type of return, which might involve changing it
+# TODO: check for duplicates and inconsistencies after the parsing
+def parse_constraints(constraints_str: str) -> list:
     """
     Parses a string of constraints into a constraint structure.
+
+    @param constraints_str: The textual representation of the list of constraints to be
+        parsed.
+    @return: The parsed constraints.
     """
 
     def _assert_valid_name(value_name):
@@ -60,7 +60,8 @@ def parse_constraints(constraints_str):
 
     # Obtain all constraints and check for disjunctions
     constraints = []
-    for constr_str in _split_fvalues(constraints_str):
+    for constr_str in split_fvalues_str(constraints_str):
+        # Collect each constraint group
         constr_group = []
         for constr in constr_str.split("|"):
             if constr[0] in "-!":
@@ -73,35 +74,22 @@ def parse_constraints(constraints_str):
                 _assert_valid_name(constr)
                 constr_group.append({"type": "presence", "fvalue": constr})
 
-        # Collect constraint group
         constraints.append(constr_group)
 
-    # Check for duplicates/inconsistencies
-    # TODO: we can have a simple check for duplicates when groups have only
-    # one entry, but it would be difficult to check when allowing disjunction
-
     return constraints
 
 
-def _split_fvalues(fvalues):
+# TODO: write test with all delimiters
+def split_fvalues_str(fvalues: str) -> list:
     """
     Split a string with multiple feature values or constraints.
 
-    This function, intended for internal usage, allows to use different
-    delimiters and guarantees that all methods will allow all delimiters.
+    This function allows to use different delimiters and guarantees that all methods will
+    allow the same delimiters. Delimiters can be white spaces, commas, semicolons,
+    forward slashes, and the " and " substring.
 
-    Delimiters can be white spaces, commas, semicolons, forward slashes,
-    and the " and " substring.
-
-    Parameters
-    ----------
-    fvalues : str
-        The string with the list of feature values to be split.
-
-    Returns
-    -------
-    fvalue_list : list
-        A list of strings with the feature values.
+    @param fvalues: The string with the list of feature values to be split.
+    @return: A list of strings with the feature values.
     """
 
     # We internally convert everything to spaces
@@ -113,20 +101,13 @@ def _split_fvalues(fvalues):
     return fvalues.split()
 
 
-def codepoint2glyph(codepoint):
+def codepoint2glyph(codepoint: str) -> str:
     """
     Convert a Unicode codepoint, given as a string, to its glyph.
 
-    Parameters
-    ----------
-    codepoint : str
-        A string with the codepoint in the format "U+XXXX", such
-        as "U+0283".
-
-    Returns
-    -------
-    glyph : str
-        The correponding glyph (such as "ʃ").
+    @param codepoint: A string with the codepoint in the format "U+XXXX", such as
+        "U+0283".
+    @return: The correponding glyph (such as "ʃ").
     """
 
     codepoint = codepoint.lower()
@@ -138,22 +119,15 @@ def codepoint2glyph(codepoint):
     return chr(value)
 
 
-def replace_codepoints(text):
+def replace_codepoints(text: str) -> str:
     """
     Replaces Unicode codepoints in a string with the corresponding glyphs.
 
     Multiple codepoints can be specified in the same string. Characters
     which are not part of codepoint annotation are kept.
 
-    Parameters
-    ----------
-    text : str
-        The text with codepoint annotations to be replaced.
-
-    Returns
-    -------
-    ret : str
-        The text with codepoint annotations replaced.
+    @param text: The text with codepoint annotations to be replaced.
+    @return: The text with codepoint annotations replaced.
     """
 
     # Internal function used for calling codepoint2glyph() on a match
@@ -163,21 +137,15 @@ def replace_codepoints(text):
     return RE_CODEPOINT.sub(_match_repr, text)
 
 
-def read_distance_matrix(filepath=None):
+# TODO: extend return annotation
+def read_distance_matrix(filepath: Optional[str] = None) -> dict:
     """
     Read a distance matrix, used to seed a regressor.
 
-    Parameters
-    ==========
-    filepath : str
-        Path to the TSV file containing the distance matrix used to
-        seed the regressor. If not provided, will default to one derived
-        from data presented in Mielke (2012) and included in the library.
-
-    Returns
-    =======
-    matrix : dict
-        A dictionary of dictionaries with the distances as floating-point
+    @param filepath: Path to the TSV file containing the distance matrix used to seed the
+        regressor. If not provided, will default to one derived from data presented in
+        Mielke (2012) and included with the library.
+    @return: A dictionary of dictionaries with the distances as floating-point
         values.
     """
 
@@ -186,7 +154,7 @@ def read_distance_matrix(filepath=None):
         filepath = filepath.as_posix()
 
     matrix = {}
-    with open(filepath) as tsvfile:
+    with open(filepath, encoding="utf-8") as tsvfile:
         for row in csv.DictReader(tsvfile, delimiter="\t"):
             grapheme = row.pop("GRAPHEME")
             matrix[grapheme] = {gr: float(dist) for gr, dist in row.items()}
@@ -194,7 +162,7 @@ def read_distance_matrix(filepath=None):
     return matrix
 
 
-def match_initial(string, candidates):
+def match_initial(string: str, candidates: List[str]) -> Tuple[str, Optional[str]]:
     """
     Returns the longest match at the initial position among a list of candidates.
 
@@ -202,22 +170,14 @@ def match_initial(string, candidates):
     of the string, making sure to match longer candidates first (so that it will
     match "abc" before "ab").
 
-    Parameters
-    ----------
-    string : str
-        The string to matched at the beginning.
-    candidates : list of str
-        A list of string candidates for initial match. The list does not need to
-        be sorted in any way.
-
-    Return
-    ------
-    string : str
-        The original string stripped of the initial match, if any. If no candidate
-        matched the beginning of the string, it will be a copy of the original one.
-    cand : str or None
-        The candidate that was matched at the beginning of the string, or `None` if
-        no match could be found.
+    @param string: The string to matched at the beginning.
+    @param candidates: A list of string candidates for initial match. The list does not
+        need to be sorted in any way.
+    @return: A tuple, whose first element is the original string stripped of the initial
+        match, if any (if no candidate matched the beginning of the string, it will be
+        a copy of the original one). The second element of the tuple is the candidate
+        that was matched at the beginning of the string, or `None` if no match could be
+        found.
     """
 
     # Sort the candidates by inverse length -- this is a bit expensive computationally,
